from jambo.parser._type_parser import GenericTypeParser
from jambo.utils.properties_builder.numeric_properties_builder import numeric_properties_builder


class IntTypeParser(GenericTypeParser):
    mapped_type = int

    json_schema_type = "integer"

    @staticmethod
    def from_properties(name, properties):
        return int, numeric_properties_builder(properties)



from jambo.parser import GenericTypeParser

from jsonschema.exceptions import SchemaError
from jsonschema.protocols import Validator
from pydantic import create_model
from pydantic.fields import Field

from typing import Type

from jambo.types.json_schema_type import JSONSchema


class SchemaConverter:
    """
    Converts JSON Schema to Pydantic models.

    This class is responsible for converting JSON Schema definitions into Pydantic models.
    It validates the schema and generates the corresponding Pydantic model with appropriate
    fields and types. The generated model can be used for data validation and serialization.
    """

    @staticmethod
    def build(schema: JSONSchema) -> Type:
        """
        Converts a JSON Schema to a Pydantic model.
        :param schema: The JSON Schema to convert.
        :return: A Pydantic model class.
        """
        if "title" not in schema:
            raise ValueError("JSON Schema must have a title.")

        return SchemaConverter.build_object(schema["title"], schema)

    @staticmethod
    def build_object(
        name: str,
        schema: JSONSchema,
    ) -> Type:
        """
        Converts a JSON Schema object to a Pydantic model given a name.
        :param name:
        :param schema:
        :return:
        """

        try:
            Validator.check_schema(schema)
        except SchemaError as e:
            raise ValueError(f"Invalid JSON Schema: {e}")

        if schema["type"] != "object":
            raise TypeError(
                f"Invalid JSON Schema: {schema['type']}. Only 'object' can be converted to Pydantic models."
            )

        return SchemaConverter._build_model_from_properties(
            name, schema["properties"], schema.get("required", [])
        )

    @staticmethod
    def _build_model_from_properties(
        model_name: str, model_properties: dict, required_keys: list[str]
    ) -> Type:
        properties = SchemaConverter._parse_properties(model_properties, required_keys)

        return create_model(model_name, **properties)

    @staticmethod
    def _parse_properties(
        properties: dict, required_keys=None
    ) -> dict[str, tuple[type, Field]]:
        required_keys = required_keys or []

        fields = {}
        for name, prop in properties.items():
            fields[name] = SchemaConverter._build_field(name, prop, required_keys)

        return fields

    @staticmethod
    def _build_field(
        name, properties: dict, required_keys: list[str]
    ) -> tuple[type, dict]:
        _field_type, _field_args = GenericTypeParser.get_impl(
            properties["type"]
        ).from_properties(name, properties)

        _field_args = _field_args or {}

        if description := properties.get("description"):
            _field_args["description"] = description

        if name not in required_keys:
            _field_args["default"] = properties.get("default", None)

        if "default_factory" in _field_args and "default" in _field_args:
            del _field_args["default"]

        return _field_type, Field(**_field_args)



from jambo.schema_converter import SchemaConverter

from pydantic import BaseModel

from unittest import TestCase


def is_pydantic_model(cls):
    return isinstance(cls, type) and issubclass(cls, BaseModel)


class TestSchemaConverter(TestCase):
    def test_jsonschema_to_pydantic(self):
        schema = {
            "title": "Person",
            "description": "A person",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
            },
            "required": ["name"],
        }

        model = SchemaConverter.build(schema)

        self.assertTrue(is_pydantic_model(model))

    def test_validation_string(self):
        schema = {
            "title": "Person",
            "description": "A person",
            "type": "object",
            "properties": {
                "name": {"type": "string", "maxLength": 4, "minLength": 1},
                "email": {
                    "type": "string",
                    "maxLength": 50,
                    "minLength": 5,
                    "pattern": r"^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$",
                },
            },
            "required": ["name"],
        }

        model = SchemaConverter.build(schema)

        self.assertEqual(model(name="John", age=30).name, "John")

        with self.assertRaises(ValueError):
            model(name=123, age=30, email="teste@hideyoshi.com")

        with self.assertRaises(ValueError):
            model(name="John Invalid", age=45, email="teste@hideyoshi.com")

        with self.assertRaises(ValueError):
            model(name="", age=45, email="teste@hideyoshi.com")

        with self.assertRaises(ValueError):
            model(name="John", age=45, email="hideyoshi.com")

    def test_validation_integer(self):
        schema = {
            "title": "Person",
            "description": "A person",
            "type": "object",
            "properties": {
                "age": {
                    "type": "integer",
                    "minimum": 0,
                    "maximum": 120,
                },
            },
            "required": ["age"],
        }

        model = SchemaConverter.build(schema)

        self.assertEqual(model(age=30).age, 30)

        with self.assertRaises(ValueError):
            model(age=-1)

        with self.assertRaises(ValueError):
            model(age=121)

    def test_validation_float(self):
        schema = {
            "title": "Person",
            "description": "A person",
            "type": "object",
            "properties": {
                "age": {
                    "type": "number",
                    "minimum": 0,
                    "maximum": 120,
                },
            },
            "required": ["age"],
        }

        model = SchemaConverter.build(schema)

        self.assertEqual(model(age=30).age, 30.0)

        with self.assertRaises(ValueError):
            model(age=-1.0)

        with self.assertRaises(ValueError):
            model(age=121.0)

    def test_validation_boolean(self):
        schema = {
            "title": "Person",
            "description": "A person",
            "type": "object",
            "properties": {
                "is_active": {"type": "boolean"},
            },
            "required": ["is_active"],
        }

        model = SchemaConverter.build(schema)

        self.assertEqual(model(is_active=True).is_active, True)

        self.assertEqual(model(is_active="true").is_active, True)

    def test_validation_list(self):
        schema = {
            "title": "Person",
            "description": "A person",
            "type": "object",
            "properties": {
                "friends": {
                    "type": "array",
                    "items": {"type": "string"},
                    "minItems": 1,
                    "maxItems": 2,
                    "uniqueItems": True,
                },
            },
            "required": ["friends"],
        }

        model = SchemaConverter.build(schema)

        self.assertEqual(
            model(friends=["John", "Jane", "John"]).friends, {"John", "Jane"}
        )

        with self.assertRaises(ValueError):
            model(friends=[])

        with self.assertRaises(ValueError):
            model(friends=["John", "Jane", "Invalid"])

    def test_validation_object(self):
        schema = {
            "title": "Person",
            "description": "A person",
            "type": "object",
            "properties": {
                "address": {
                    "type": "object",
                    "properties": {
                        "street": {"type": "string"},
                        "city": {"type": "string"},
                    },
                    "required": ["street", "city"],
                },
            },
            "required": ["address"],
        }

        model = SchemaConverter.build(schema)

        obj = model(address={"street": "123 Main St", "city": "Springfield"})

        self.assertEqual(obj.address.street, "123 Main St")
        self.assertEqual(obj.address.city, "Springfield")

    def test_default_for_string(self):
        schema = {
            "title": "Person",
            "description": "A person",
            "type": "object",
            "properties": {
                "name": {
                    "type": "string",
                    "default": "John",
                },
            },
            "required": ["name"],
        }

        model = SchemaConverter.build(schema)

        obj = model(name="John")

        self.assertEqual(obj.name, "John")

        # Test for default with maxLength
        schema_max_length = {
            "title": "Person",
            "description": "A person",
            "type": "object",
            "properties": {
                "name": {
                    "type": "string",
                    "default": "John",
                    "maxLength": 2,
                },
            },
            "required": ["name"],
        }

        with self.assertRaises(ValueError):
            SchemaConverter.build(schema_max_length)

    def test_default_for_list(self):
        schema_list = {
            "title": "Person",
            "description": "A person",
            "type": "object",
            "properties": {
                "friends": {
                    "type": "array",
                    "items": {"type": "string"},
                    "default": ["John", "Jane"],
                },
            },
            "required": ["friends"],
        }

        model_list = SchemaConverter.build(schema_list)

        self.assertEqual(model_list().friends, ["John", "Jane"])

        # Test for default with uniqueItems
        schema_set = {
            "title": "Person",
            "description": "A person",
            "type": "object",
            "properties": {
                "friends": {
                    "type": "array",
                    "items": {"type": "string"},
                    "default": ["John", "Jane"],
                    "uniqueItems": True,
                },
            },
            "required": ["friends"],
        }

        model_set = SchemaConverter.build(schema_set)

        self.assertEqual(model_set().friends, {"John", "Jane"})



from jambo.parser import (
    ArrayTypeParser,
    FloatTypeParser,
    GenericTypeParser,
    IntTypeParser,
    ObjectTypeParser,
    StringTypeParser,
)

import unittest
from typing import get_args


class TestTypeParser(unittest.TestCase):
    def test_get_impl(self):
        self.assertEqual(GenericTypeParser.get_impl("integer"), IntTypeParser)
        self.assertEqual(GenericTypeParser.get_impl("string"), StringTypeParser)
        self.assertEqual(GenericTypeParser.get_impl("number"), FloatTypeParser)
        self.assertEqual(GenericTypeParser.get_impl("object"), ObjectTypeParser)
        self.assertEqual(GenericTypeParser.get_impl("array"), ArrayTypeParser)

    def test_int_parser(self):
        parser = IntTypeParser()

        type_parsing, type_validator = parser.from_properties(
            "placeholder",
            {
                "type": "integer",
                "minimum": 0,
                "exclusiveMinimum": 1,
                "maximum": 10,
                "exclusiveMaximum": 11,
                "multipleOf": 2,
            },
        )

        self.assertEqual(type_parsing, int)
        self.assertEqual(type_validator["ge"], 0)
        self.assertEqual(type_validator["gt"], 1)
        self.assertEqual(type_validator["le"], 10)
        self.assertEqual(type_validator["lt"], 11)
        self.assertEqual(type_validator["multiple_of"], 2)

    def test_float_parser(self):
        parser = FloatTypeParser()

        type_parsing, type_validator = parser.from_properties(
            "placeholder",
            {
                "type": "number",
                "minimum": 0,
                "exclusiveMinimum": 1,
                "maximum": 10,
                "exclusiveMaximum": 11,
                "multipleOf": 2,
            },
        )

        self.assertEqual(type_parsing, float)
        self.assertEqual(type_validator["ge"], 0)
        self.assertEqual(type_validator["gt"], 1)
        self.assertEqual(type_validator["le"], 10)
        self.assertEqual(type_validator["lt"], 11)
        self.assertEqual(type_validator["multiple_of"], 2)

    def test_string_parser(self):
        parser = StringTypeParser()

        type_parsing, type_validator = parser.from_properties(
            "placeholder",
            {
                "type": "string",
                "maxLength": 10,
                "minLength": 1,
                "pattern": "[a-zA-Z0-9]",
            },
        )

        self.assertEqual(type_parsing, str)
        self.assertEqual(type_validator["max_length"], 10)
        self.assertEqual(type_validator["min_length"], 1)
        self.assertEqual(type_validator["pattern"], "[a-zA-Z0-9]")

    def test_object_parser(self):
        parser = ObjectTypeParser()

        properties = {
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
            },
        }

        Model, _args = parser.from_properties("placeholder", properties)

        obj = Model(name="name", age=10)

        self.assertEqual(obj.name, "name")
        self.assertEqual(obj.age, 10)

    def test_array_of_string_parser(self):
        parser = ArrayTypeParser()
        expected_definition = (list[str], {})

        properties = {"items": {"type": "string"}}

        self.assertEqual(
            parser.from_properties("placeholder", properties), expected_definition
        )

    def test_array_of_object_parser(self):
        parser = ArrayTypeParser()

        properties = {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "name": {"type": "string"},
                    "age": {"type": "integer"},
                },
            },
            "maxItems": 10,
            "minItems": 1,
            "uniqueItems": True,
        }

        type_parsing, type_validator = parser.from_properties("placeholder", properties)

        self.assertEqual(type_parsing.__origin__, set)
        self.assertEqual(type_validator["max_length"], 10)
        self.assertEqual(type_validator["min_length"], 1)

        Model = get_args(type_parsing)[0]
        obj = Model(name="name", age=10)

        self.assertEqual(obj.name, "name")
        self.assertEqual(obj.age, 10)



from abc import ABC, abstractmethod
from typing import Generic, TypeVar
from typing_extensions import Self

from pydantic import Field

T = TypeVar("T")


class GenericTypeParser(ABC, Generic[T]):
    @property
    @abstractmethod
    def mapped_type(self) -> type[T]: ...

    @property
    @abstractmethod
    def json_schema_type(self) -> str: ...

    @staticmethod
    @abstractmethod
    def from_properties(
        name: str, properties: dict[str, any]
    ) -> tuple[type[T], Field]: ...

    @classmethod
    def get_impl(cls, type_name: str) -> Self:
        for subcls in cls.__subclasses__():
            if subcls.json_schema_type == type_name:
                return subcls

        raise ValueError(f"Unknown type: {type_name}")



# Exports generic type parser
from ._type_parser import GenericTypeParser as GenericTypeParser

# Exports Implementations
from .int_type_parser import IntTypeParser as IntTypeParser
from .object_type_parser import ObjectTypeParser as ObjectTypeParser
from .string_type_parser import StringTypeParser as StringTypeParser
from .array_type_parser import ArrayTypeParser as ArrayTypeParser
from .boolean_type_parser import BooleanTypeParser as BooleanTypeParser
from .float_type_parser import FloatTypeParser as FloatTypeParser



def mappings_properties_builder(properties, mappings, default_mappings=None):
    default_mappings = default_mappings or {
        "default": "default",
        "description": "description",
    }

    mappings = default_mappings | mappings

    return {
        mappings[key]: value for key, value in properties.items() if key in mappings
    }



from jambo.parser._type_parser import GenericTypeParser
from jambo.utils.properties_builder.mappings_properties_builder import (
    mappings_properties_builder,
)


class BooleanTypeParser(GenericTypeParser):
    mapped_type = bool

    json_schema_type = "boolean"

    @staticmethod
    def from_properties(name, properties):
        _mappings = {
            "default": "default",
        }
        return bool, mappings_properties_builder(properties, _mappings)



from jambo.parser._type_parser import GenericTypeParser
from jambo.utils.properties_builder.numeric_properties_builder import numeric_properties_builder


class FloatTypeParser(GenericTypeParser):
    mapped_type = float

    json_schema_type = "number"

    @staticmethod
    def from_properties(name, properties):
        return float, numeric_properties_builder(properties)



from jambo.utils.properties_builder.mappings_properties_builder import (
    mappings_properties_builder,
)


def numeric_properties_builder(properties):
    _mappings = {
        "minimum": "ge",
        "exclusiveMinimum": "gt",
        "maximum": "le",
        "exclusiveMaximum": "lt",
        "multipleOf": "multiple_of",
        "default": "default",
    }

    mapped_properties = mappings_properties_builder(properties, _mappings)

    if "default" in properties:
        default_value = properties["default"]
        if not isinstance(default_value, (int, float)):
            raise ValueError(
                f"Default value must be a number, got {type(default_value).__name__}"
            )

        if default_value >= properties.get("maximum", float("inf")):
            raise ValueError(
                f"Default value exceeds maximum limit of {properties.get('maximum')}"
            )

        if default_value <= properties.get("minimum", float("-inf")):
            raise ValueError(
                f"Default value is below minimum limit of {properties.get('minimum')}"
            )

        if default_value > properties.get("exclusiveMaximum", float("inf")):
            raise ValueError(
                f"Default value exceeds exclusive maximum limit of {properties.get('exclusiveMaximum')}"
            )

        if default_value < properties.get("exclusiveMinimum", float("-inf")):
            raise ValueError(
                f"Default value is below exclusive minimum limit of {properties.get('exclusiveMinimum')}"
            )

        if "multipleOf" in properties:
            if default_value % properties["multipleOf"] != 0:
                raise ValueError(
                    f"Default value {default_value} is not a multiple of {properties['multipleOf']}"
                )

    return mapped_properties



import copy

from jambo.parser._type_parser import GenericTypeParser

from typing import TypeVar

from jambo.utils.properties_builder.mappings_properties_builder import (
    mappings_properties_builder,
)

V = TypeVar("V")


class ArrayTypeParser(GenericTypeParser):
    mapped_type = list

    json_schema_type = "array"

    @classmethod
    def from_properties(cls, name, properties):
        _item_type, _item_args = GenericTypeParser.get_impl(
            properties["items"]["type"]
        ).from_properties(name, properties["items"])

        _mappings = {
            "maxItems": "max_length",
            "minItems": "min_length",
        }

        wrapper_type = set if properties.get("uniqueItems", False) else list

        mapped_properties = mappings_properties_builder(
            properties, _mappings, {"description": "description"}
        )

        if "default" in properties:
            default_list = properties["default"]
            if not isinstance(default_list, list):
                raise ValueError(
                    f"Default value must be a list, got {type(default_list).__name__}"
                )

            if len(default_list) > properties.get("maxItems", float("inf")):
                raise ValueError(
                    f"Default list exceeds maxItems limit of {properties.get('maxItems')}"
                )

            if len(default_list) < properties.get("minItems", 0):
                raise ValueError(
                    f"Default list is below minItems limit of {properties.get('minItems')}"
                )

            if not all(isinstance(item, _item_type) for item in default_list):
                raise ValueError(
                    f"All items in the default list must be of type {_item_type.__name__}"
                )

            if wrapper_type is list:
                mapped_properties["default_factory"] = lambda: copy.deepcopy(
                    wrapper_type(default_list)
                )
            else:
                mapped_properties["default_factory"] = lambda: wrapper_type(
                    default_list
                )

        return wrapper_type[_item_type], mapped_properties



from jambo.parser._type_parser import GenericTypeParser
from jambo.utils.properties_builder.mappings_properties_builder import (
    mappings_properties_builder,
)


class StringTypeParser(GenericTypeParser):
    mapped_type = str

    json_schema_type = "string"

    @staticmethod
    def from_properties(name, properties):
        _mappings = {
            "maxLength": "max_length",
            "minLength": "min_length",
            "pattern": "pattern",
        }

        mapped_properties = mappings_properties_builder(properties, _mappings)

        if "default" in properties:
            default_value = properties["default"]
            if not isinstance(default_value, str):
                raise ValueError(
                    f"Default value for {name} must be a string, "
                    f"but got {type(properties['default'])}."
                )

            if len(default_value) > properties.get("maxLength", float("inf")):
                raise ValueError(
                    f"Default value for {name} exceeds maxLength limit of {properties.get('maxLength')}"
                )

            if len(default_value) < properties.get("minLength", 0):
                raise ValueError(
                    f"Default value for {name} is below minLength limit of {properties.get('minLength')}"
                )

        return str, mapped_properties



from typing import List, Dict, Union, TypedDict, Literal


JSONSchemaType = Literal[
    "string", "number", "integer", "boolean", "object", "array", "null"
]


JSONType = Union[str, int, float, bool, None, Dict[str, "JSONType"], List["JSONType"]]


class JSONSchema(TypedDict, total=False):
    # Basic metadata
    title: str
    description: str
    default: JSONType
    examples: List[JSONType]

    # Type definitions
    type: Union[JSONSchemaType, List[JSONSchemaType]]

    # Object-specific keywords
    properties: Dict[str, "JSONSchema"]
    required: List[str]
    additionalProperties: Union[bool, "JSONSchema"]
    minProperties: int
    maxProperties: int
    patternProperties: Dict[str, "JSONSchema"]
    dependencies: Dict[str, Union[List[str], "JSONSchema"]]

    # Array-specific keywords
    items: Union["JSONSchema", List["JSONSchema"]]
    additionalItems: Union[bool, "JSONSchema"]
    minItems: int
    maxItems: int
    uniqueItems: bool

    # String-specific keywords
    minLength: int
    maxLength: int
    pattern: str
    format: str

    # Number-specific keywords
    minimum: float
    maximum: float
    exclusiveMinimum: float
    exclusiveMaximum: float
    multipleOf: float

    # Enum and const
    enum: List[JSONType]
    const: JSONType

    # Conditionals
    if_: "JSONSchema"  # 'if' is a reserved word in Python
    then: "JSONSchema"
    else_: "JSONSchema"  # 'else' is also a reserved word

    # Combination keywords
    allOf: List["JSONSchema"]
    anyOf: List["JSONSchema"]
    oneOf: List["JSONSchema"]
    not_: "JSONSchema"  # 'not' is a reserved word


# Fix forward references
JSONSchema.__annotations__["properties"] = Dict[str, JSONSchema]
JSONSchema.__annotations__["items"] = Union[JSONSchema, List[JSONSchema]]
JSONSchema.__annotations__["additionalItems"] = Union[bool, JSONSchema]
JSONSchema.__annotations__["additionalProperties"] = Union[bool, JSONSchema]
JSONSchema.__annotations__["patternProperties"] = Dict[str, JSONSchema]
JSONSchema.__annotations__["dependencies"] = Dict[str, Union[List[str], JSONSchema]]
JSONSchema.__annotations__["if_"] = JSONSchema
JSONSchema.__annotations__["then"] = JSONSchema
JSONSchema.__annotations__["else_"] = JSONSchema
JSONSchema.__annotations__["allOf"] = List[JSONSchema]
JSONSchema.__annotations__["anyOf"] = List[JSONSchema]
JSONSchema.__annotations__["oneOf"] = List[JSONSchema]
JSONSchema.__annotations__["not_"] = JSONSchema



from jambo.parser._type_parser import GenericTypeParser


class ObjectTypeParser(GenericTypeParser):
    mapped_type = object

    json_schema_type = "object"

    @staticmethod
    def from_properties(name, properties):
        from jambo.schema_converter import SchemaConverter

        if "default" in properties:
            raise RuntimeError("Default values for objects are not supported.")

        return (
            SchemaConverter.build_object(name, properties),
            {},  # The second argument is not used in this case
        )
